from __future__ import annotations

from dataclasses import dataclass, field
from typing import (IO, Any, Callable, Generator, Iterable, TypeAlias, TypeVar,
                    ParamSpec)

from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import OperationInvT, SSAValue, Operation
from xdsl.utils.exceptions import InterpretationError


@dataclass
class InterpreterFunctions:
    """
    Holds the Python implementations for Operations. Users should
    subclass this class, and define the functions to run during interpretation.
    For example:

    ``` python
    @register_impls
    class ArithFunctions(InterpreterFunctions):

        @impl(arith.Addi)
        def run_addi(self, interpreter: Interpreter, op: arith.Addi, 
                     args: tuple[Any, ...]) -> tuple[Any, ...]:
            lhs, rhs = args
            return lhs + rhs, 
    ```

    The interpreter will take care of fetching the Python values associated 
    with the operand SSAValues, and setting the return values to the 
    appropriate OpResults.

    To override the definition of an operation implementation, subclass the
    class to override, and redefine the functions, annotating them with 
    `@impl`.

    ``` python
    @register_impls
    class DebugArithFunctions(ArithFunctions):

        @impl(arith.Addi)
        def run_addi(self, interpreter: Interpreter, op: arith.Addi, 
                     args: tuple[Any, ...]) -> tuple[Any, ...]:
            lhs, rhs = args
            print(lhs, rhs, lhs + rhs)
            return lhs + rhs, 
    ```
    """

    @classmethod
    def _impls(
        cls
    ) -> Iterable[tuple[type[Operation], OpImpl[InterpreterFunctions,
                                                Operation]]]:
        try:
            impl_dict = getattr(cls, _IMPL_DICT)
            return impl_dict.items()
        except AttributeError as e:
            raise ValueError(
                f'Use `@register_impls` on class {cls.__name__}') from e


_FT = TypeVar('_FT', bound=InterpreterFunctions)

_IMPL_OP_TYPE = '__impl_op_type'
_IMPL_DICT = '__impl_dict'

P = ParamSpec('P')


def impl(
    op_type: type[OperationInvT]
) -> Callable[[OpImpl[_FT, OperationInvT]], OpImpl[_FT, OperationInvT]]:
    """
    Marks the Python implementation of an xDSL `Operation` instance, to be used
    by an `Interpreter`. The Interpreter will fetch the Python values 
    associated with the operands from the current environment, and pass them as
    the `args` parameter. The returned values are assigned to the `results` 
    values.
    
    See `InterpreterFunctions`
    """

    def annot(func: OpImpl[_FT, OperationInvT]) -> OpImpl[_FT, OperationInvT]:
        setattr(func, _IMPL_OP_TYPE, op_type)
        return func

    return annot


def register_impls(ft: type[_FT]) -> type[_FT]:
    """
    Enumerates the methods on a given class, and registers the ones marked with
    `@impl` in a way that an `Interpreter` instance can find them for dynamic 
    dispatch during interpretation.
    
    See `InterpreterFunctions`
    """
    impl_dict: _ImplDict = {}
    for cls in ft.mro():
        # Iterate from subclass through superclasses
        # Assign definitions, unless they've been redefined in a subclass
        for val in cls.__dict__.values():
            if _IMPL_OP_TYPE in val.__dir__():
                # This is an annotated function
                op_type = getattr(val, _IMPL_OP_TYPE)
                if op_type not in impl_dict:
                    # subclass overrides superclass definition
                    impl_dict[op_type] = val  # type: ignore
    setattr(ft, _IMPL_DICT, impl_dict)

    return ft


@dataclass
class _InterpreterFunctionImpls:
    """
    Used to combine multiple function implementations. The operation
    implementations need to be passed the instance of the Functions class,
    so we keep a `(Functions, OpImpl)` tuple for every Operation type.
    """

    _impl_dict: dict[type[Operation],
                     tuple[InterpreterFunctions,
                           OpImpl[InterpreterFunctions,
                                  Operation]]] = field(default_factory=dict)

    def register_from(self, ft: InterpreterFunctions, /, override: bool):
        impls = ft._impls()  # pyright: ignore[reportPrivateUsage]
        for op_type, impl in impls:
            if op_type in self._impl_dict and not override:
                raise ValueError(
                    "Attempting to register implementation for op of type "
                    f"{op_type}, but type already registered")

            self._impl_dict[op_type] = (ft, impl)

    def run(self, interpreter: Interpreter, op: Operation,
            args: tuple[Any, ...]) -> tuple[Any, ...]:
        if type(op) not in self._impl_dict:
            raise InterpretationError(
                f'Could not find interpretation function for op {op.name}')
        ft, impl = self._impl_dict[type(op)]
        return impl(ft, interpreter, op, args)


@dataclass
class InterpreterContext:
    """
    Class holding the Python values associated with SSAValues during an
    interpretation context. An environment is a stack of scopes, values are
    assigned to the current scope, but can be fetched from a parent scope.
    """

    name: str = field(default="unknown")
    parent: InterpreterContext | None = None
    env: dict[SSAValue, Any] = field(default_factory=dict)

    def __getitem__(self, key: SSAValue) -> Any:
        """
        Fetch key from environment. Attempts to first fetch from current scope,
        then from parent scopes. Raises Interpretation error if not found.
        """
        if key in self.env:
            return self.env[key]
        if self.parent is not None:
            return self.parent[key]
        raise InterpretationError(f'Could not find value for {key} in {self}')

    def __setitem__(self, key: SSAValue, value: Any):
        """
        Assign key to current scope. Raises InterpretationError if key already
        assigned to.
        """
        if key in self.env:
            raise InterpretationError(
                f'Attempting to register SSAValue {value} for name {key}'
                f', but value with that name already exists in {self}')
        self.env[key] = value

    def stack(self) -> Generator[InterpreterContext, None, None]:
        """
        Iterates through scopes starting with the root scope.
        """
        if self.parent is not None:
            yield from self.parent.stack()
        yield self

    def __format__(self, __format_spec: str) -> str:
        return '/'.join(c.name for c in self.stack())


@dataclass
class Interpreter:
    """
    An extensible interpreter, initialised with a Module to interpret. The 
    implementation for each Operation subclass should be provided via a 
    `InterpretationFunctions` instance. Interpretations can be overridden, and 
    the override must be specified explicitly, by passing `override=True` to
    the `register_functions` method.
    """

    module: ModuleOp
    _impls: _InterpreterFunctionImpls = field(
        default_factory=_InterpreterFunctionImpls)
    _ctx: InterpreterContext = field(
        default_factory=lambda: InterpreterContext(name='root'))
    file: IO[str] | None = field(default=None)

    def get_values(self, values: Iterable[SSAValue]) -> tuple[Any, ...]:
        """
        Get values from current environment.
        """
        return tuple(self._ctx[value] for value in values)

    def set_values(self, pairs: Iterable[tuple[SSAValue, Any]]):
        """
        Set values to current scope.
        Raises InterpretationError if len(ssa_values) != len(result_values), or
        if SSA value already has a Python value in the current scope.
        """
        for ssa_value, result_value in pairs:
            self._ctx[ssa_value] = result_value

    def push_scope(self, name: str = 'unknown') -> None:
        """
        Create new scope in current environment, with optional custom `name`.
        """
        self._ctx = InterpreterContext(name, self._ctx)

    def pop_scope(self) -> None:
        """
        Discard the current scope, and all the values registered in it. Sets 
        parent scope of current scope to new current scope.
        Raises InterpretationError if current scope is root scope.
        """
        if self._ctx.parent is None:
            raise InterpretationError('Attempting to pop root env')

        self._ctx = self._ctx.parent

    def register_implementations(self,
                                 impls: InterpreterFunctions,
                                 /,
                                 override: bool = False) -> None:
        """
        Register implementations for operations defined in given 
        `InterpreterFunctions` object. Raise InterpretationError if an 
        operation already has an implementation registered, unless override is 
        set to True.
        """
        self._impls.register_from(impls, override=override)

    def run(self, op: Operation):
        """
        Fetches the implemetation for the given op, passes it the Python values
        associated with the SSA operands, and assigns the results to the 
        operation's results.
        """
        inputs = self.get_values(op.operands)
        results = self._impls.run(self, op, inputs)
        self.interpreter_assert(
            len(op.results) == len(results), 'Incorrect number of results')
        self.set_values(zip(op.results, results))

    def run_module(self):
        """Starts execution of `self.module`"""
        self.run(self.module)

    def print(self, *args: Any, **kwargs: Any):
        """Print to current file."""
        print(*args, **kwargs, file=self.file)

    def interpreter_assert(self, condition: bool, message: str | None = None):
        """Raise InterpretationError if condition is not satisfied."""
        if not condition:
            raise InterpretationError(
                f'AssertionError: ({self._ctx})({message})')


OpImpl: TypeAlias = Callable[
    [_FT, Interpreter, OperationInvT, tuple[Any, ...]], tuple[Any, ...]]

_ImplDict: TypeAlias = dict[type[Operation], OpImpl[InterpreterFunctions,
                                                    Operation]]



from __future__ import annotations
from typing import TypeVar, Iterable, SupportsIndex, List

_T = TypeVar('_T')


class IList(List[_T]):
    """
    A list that can be frozen. Once frozen, it can not be modified. 
    In comparison to FrozenList this supports pattern matching.
    """
    _frozen: bool = False

    def freeze(self):
        self._frozen = True

    def _unfreeze(self):
        self._frozen = False

    def append(self, __object: _T) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().append(__object)

    def extend(self, __iterable: Iterable[_T]) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().extend(__iterable)

    def insert(self, __index: SupportsIndex, __object: _T) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().insert(__index, __object)

    def remove(self, __value: _T) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().remove(__value)

    def pop(self, __index: SupportsIndex = ...) -> _T:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().pop(__index)

    def clear(self) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().clear()

    def __setitem__(self, __index: SupportsIndex, __object: _T) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().__setitem__(__index, __object)

    def __delitem__(self, __index: SupportsIndex | slice) -> None:
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().__delitem__(__index)

    def __add__(self, __x: Iterable[_T]) -> IList[_T]:
        return IList(super().__add__(__x))  # type: ignore

    def __iadd__(self, __x: Iterable[_T]):
        if self._frozen:
            raise Exception("frozen list can not be modified")
        return super().__iadd__(__x)  # type: ignore

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, IList):
            return super().__eq__(__o)  # type: ignore
        return False



import xdsl.dialects.arith as arith

from typing import Callable, TypeVar, Union
from xdsl.frontend.dialects.builtin import index, i1, i32, i64, f16, f32, f64
from xdsl.ir import Operation

_Int = TypeVar("_Int", bound=Union[index, i1, i32, i64])


def addi(lhs: _Int, rhs: _Int) -> _Int:
    ...


def resolve_addi() -> Callable[..., Operation]:
    return arith.Addi.get


def andi(lhs: _Int, rhs: _Int) -> _Int:
    ...


def resolve_andi() -> Callable[..., Operation]:
    return arith.AndI.get


def cmpi(lhs: _Int, rhs: _Int, mnemonic: str) -> i1:
    ...


def resolve_cmpi() -> Callable[..., Operation]:
    return arith.Cmpi.get


def muli(lhs: _Int, rhs: _Int) -> _Int:
    ...


def resolve_muli() -> Callable[..., Operation]:
    return arith.Muli.get


def shli(lhs: _Int, rhs: _Int) -> _Int:
    ...


def resolve_shli() -> Callable[..., Operation]:
    return arith.ShLI.get


def shrsi(lhs: _Int, rhs: _Int) -> _Int:
    ...


def resolve_shrsi() -> Callable[..., Operation]:
    return arith.ShRSI.get


def subi(lhs: _Int, rhs: _Int) -> _Int:
    ...


def resolve_subi() -> Callable[..., Operation]:
    return arith.Subi.get


_Float = TypeVar("_Float", bound=Union[f16, f32, f64])


def addf(lhs: _Float, rhs: _Float) -> _Float:
    ...


def resolve_addf() -> Callable[..., Operation]:
    return arith.Addf.get


def mulf(lhs: _Float, rhs: _Float) -> _Float:
    ...


def resolve_mulf() -> Callable[..., Operation]:
    return arith.Mulf.get


def subf(lhs: _Float, rhs: _Float) -> _Float:
    ...


def resolve_subf() -> Callable[..., Operation]:
    return arith.Subf.get



from __future__ import annotations
from types import TracebackType

from typing import Callable, ClassVar, Sequence, TypeAlias, overload
import threading
import contextlib

from dataclasses import dataclass, field
from xdsl.dialects.builtin import ArrayAttr

from xdsl.ir import Operation, OperationInvT, Attribute, Region, Block, BlockArgument


@dataclass
class Builder:
    """
    A helper class to construct IRs, by keeping track of where to insert an
    operation. Currently the insertion point is always at the end of the block.
    In the future will closely follow the API of `OpBuilder` in MLIR, inserting 
    at arbitrary locations.

    https://mlir.llvm.org/doxygen/classmlir_1_1OpBuilder.html
    """

    block: Block
    """
    Operations will be inserted in this block.
    """

    def insert(self, op: OperationInvT) -> OperationInvT:
        """
        Inserts `op` in `self.block` at the end of the block.
        """

        implicit_builder = _ImplicitBuilder.get()

        if implicit_builder is not None and implicit_builder is not self:
            raise ValueError(
                'Cannot insert operation explicitly when an implicit '
                'builder exists.')

        self.block.add_op(op)
        return op

    @staticmethod
    def _region_no_args(func: Callable[[Builder], None]) -> Region:
        """
        Generates a single-block region.
        """
        block = Block()
        builder = Builder(block)
        func(builder)
        return Region([block])

    @staticmethod
    def _region_args(
        input_types: Sequence[Attribute] | ArrayAttr[Attribute]
    ) -> Callable[[_CallableRegionFuncType], Region]:
        """
        Decorator for constructing a single-block region, containing the implementation of a
        region with some input arguments.
        """

        if isinstance(input_types, ArrayAttr):
            input_types = input_types.data

        def wrapper(func: _CallableRegionFuncType) -> Region:
            block = Block(arg_types=input_types)
            builder = Builder(block)

            func(builder, block.args)

            region = Region([block])
            return region

        return wrapper

    @overload
    @staticmethod
    def region(
        input: Sequence[Attribute] | ArrayAttr[Attribute]
    ) -> Callable[[_CallableRegionFuncType], Region]:
        """
        Annotation used to construct a Region tuple from a function.
        The annotation can be used in two ways:

        For regions that have inputs or outputs:
        ```
        @Builder.region(input_types)
        def func(builder: Builder, args: tuple[BlockArgument, ...]) -> None:
            ...
        ```

        For regions that don't have inputs or outputs:
        ``` python
        @Builder.region
        def func(builder: Builder) -> None:
            ...
        ```
        """
        ...

    @overload
    @staticmethod
    def region(input: Callable[[Builder], None]) -> Region:
        ...

    @staticmethod
    def region(
        input: Sequence[Attribute] | ArrayAttr[Attribute]
        | Callable[[Builder], None]
    ) -> Callable[[_CallableRegionFuncType], Region] | Region:
        if isinstance(input, Callable):
            return Builder._region_no_args(input)
        else:
            return Builder._region_args(input)

    @staticmethod
    def _implicit_region_no_args(func: Callable[[], None]) -> Region:
        """
        Generates a single-block region.
        """
        block = Block()
        builder = Builder(block)

        with _ImplicitBuilder(builder):
            func()

        return Region([block])

    @staticmethod
    def _implicit_region_args(
        input_types: Sequence[Attribute] | ArrayAttr[Attribute]
    ) -> Callable[[_CallableImplicitRegionFuncType], Region]:
        """
        Decorator for constructing a single-block region, containing the implementation of a
        region with some input arguments.
        """

        if isinstance(input_types, ArrayAttr):
            input_types = input_types.data

        def wrapper(func: _CallableImplicitRegionFuncType) -> Region:
            block = Block(arg_types=input_types)
            builder = Builder(block)

            with _ImplicitBuilder(builder):
                func(block.args)

            region = Region([block])
            return region

        return wrapper

    @overload
    @staticmethod
    def implicit_region(
        input: Sequence[Attribute] | ArrayAttr[Attribute]
    ) -> Callable[[_CallableImplicitRegionFuncType], Region]:
        """
        Annotation used to construct a Region tuple from a function.
        The annotation can be used in two ways:

        For regions that have inputs or outputs:
        ```
        @Builder.implicit_region(input_types)
        def func(args: tuple[BlockArgument, ...]) -> None:
            ...
        ```

        For regions that don't have inputs or outputs:
        ``` python
        @Builder.implicit_region
        def func() -> None:
            ...
        ```
        """
        ...

    @overload
    @staticmethod
    def implicit_region(input: Callable[[], None]) -> Region:
        ...

    @staticmethod
    def implicit_region(
        input: Sequence[Attribute] | ArrayAttr[Attribute]
        | Callable[[], None]
    ) -> Callable[[_CallableImplicitRegionFuncType], Region] | Region:
        if isinstance(input, Callable):
            return Builder._implicit_region_no_args(input)
        else:
            return Builder._implicit_region_args(input)


# Implicit builders


@dataclass
class _ImplicitBuilderStack(threading.local):
    """
    Stores the stack of implicit builders for use in @Builder.implicit_region, empty by
    default. There is a stack per thread, guaranteed by inheriting from `threading.local`.
    """
    stack: list[Builder] = field(default_factory=list)

    def push(self, builder: Builder) -> None:
        self.stack.append(builder)

    def get(self) -> Builder | None:
        if len(self.stack):
            return self.stack[-1]

    def pop(self, builder: Builder) -> Builder:
        popped = self.stack.pop()
        assert popped is builder
        return popped


@dataclass
class _ImplicitBuilder(contextlib.AbstractContextManager[None]):
    """
    Stores the current implicit builder context, consisting of the stack of builders in
    the current thread, and the current builder.
    """

    _stack: ClassVar[_ImplicitBuilderStack] = _ImplicitBuilderStack()

    builder: Builder

    def __enter__(self) -> None:
        type(self)._stack.push(self.builder)

    def __exit__(self, __exc_type: type[BaseException] | None,
                 __exc_value: BaseException | None,
                 __traceback: TracebackType | None) -> bool | None:
        type(self)._stack.pop(self.builder)

    @classmethod
    def get(cls) -> Builder | None:
        return cls._stack.get()


_CallableRegionFuncType: TypeAlias = Callable[
    [Builder, tuple[BlockArgument, ...]], None]
_CallableImplicitRegionFuncType: TypeAlias = Callable[
    [tuple[BlockArgument, ...]], None]


def _op_init_callback(op: Operation):
    if (b := _ImplicitBuilder.get()) is not None:
        b.insert(op)


def _override_operation_post_init() -> None:
    old_post_init = Operation.__post_init__

    def new_post_init(self: Operation) -> None:
        old_post_init(self)
        _op_init_callback(self)

    Operation.__post_init__ = new_post_init


# set up the operation callback for implicit construction
_override_operation_post_init()



from __future__ import annotations

from typing import (TYPE_CHECKING, Annotated, Sequence, TypeVar, Optional,
                    List, TypeAlias, cast)

from xdsl.dialects.builtin import (AnyIntegerAttr, DenseIntOrFPElementsAttr,
                                   IntegerAttr, DenseArrayBase, IndexType,
                                   StridedLayoutAttr, ArrayAttr, NoneAttr,
                                   SymbolRefAttr, i64, StringAttr, UnitAttr)
from xdsl.ir import (TypeAttribute, Operation, SSAValue, ParametrizedAttribute,
                     Dialect, OpResult)
from xdsl.irdl import (irdl_attr_definition, irdl_op_definition, ParameterDef,
                       Generic, Attribute, AnyAttr, Operand, VarOperand,
                       AttrSizedOperandSegments, OpAttr)
from xdsl.utils.exceptions import VerifyException

if TYPE_CHECKING:
    from xdsl.parser import BaseParser
    from xdsl.printer import Printer

_MemRefTypeElement = TypeVar("_MemRefTypeElement", bound=Attribute)


@irdl_attr_definition
class MemRefType(Generic[_MemRefTypeElement], ParametrizedAttribute,
                 TypeAttribute):
    name = "memref"

    shape: ParameterDef[ArrayAttr[AnyIntegerAttr]]
    element_type: ParameterDef[_MemRefTypeElement]
    layout: ParameterDef[Attribute]
    memory_space: ParameterDef[Attribute]

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> List[int]:
        return [i.value.data for i in self.shape.data]

    @staticmethod
    def from_element_type_and_shape(
        referenced_type: _MemRefTypeElement,
        shape: Sequence[int | AnyIntegerAttr],
        layout: Attribute = NoneAttr(),
        memory_space: Attribute = NoneAttr()
    ) -> MemRefType[_MemRefTypeElement]:
        return MemRefType([
            ArrayAttr[AnyIntegerAttr]([
                d if isinstance(d, IntegerAttr) else
                IntegerAttr.from_index_int_value(d) for d in shape
            ]), referenced_type, layout, memory_space
        ])

    @staticmethod
    def from_params(
        referenced_type: _MemRefTypeElement,
        shape: ArrayAttr[AnyIntegerAttr] = ArrayAttr(
            [IntegerAttr.from_int_and_width(1, 64)]),
        layout: Attribute = NoneAttr(),
        memory_space: Attribute = NoneAttr()
    ) -> MemRefType[_MemRefTypeElement]:
        return MemRefType([shape, referenced_type, layout, memory_space])

    @staticmethod
    def parse_parameters(parser: BaseParser) -> list[Attribute]:
        parser._synchronize_lexer_and_tokenizer(  # pyright: ignore[reportPrivateUsage]
        )
        parser.parse_punctuation('<', ' in memref attribute')
        shape = parser.parse_attribute()
        parser.parse_punctuation(',',
                                 ' between shape and element type parameters')
        type = parser.parse_attribute()
        # If we have a layout or a memory space, parse both of them.
        if parser.parse_optional_punctuation(',') is None:
            parser.parse_punctuation('>', ' at end of memref attribute')
            return [shape, type, NoneAttr(), NoneAttr()]
        layout = parser.parse_attribute()
        parser.parse_punctuation(',', ' between layout and memory space')
        memory_space = parser.parse_attribute()
        parser.parse_punctuation('>', ' at end of memref attribute')
        parser._synchronize_lexer_and_tokenizer(  # pyright: ignore[reportPrivateUsage]
        )

        return [shape, type, layout, memory_space]

    def print_parameters(self, printer: Printer) -> None:
        printer.print('<', self.shape, ', ', self.element_type)
        if self.layout != NoneAttr() or self.memory_space != NoneAttr():
            printer.print(', ', self.layout, ', ', self.memory_space)
        printer.print('>')


_UnrankedMemrefTypeElems = TypeVar("_UnrankedMemrefTypeElems",
                                   bound=Attribute,
                                   covariant=True)
_UnrankedMemrefTypeElemsInit = TypeVar("_UnrankedMemrefTypeElemsInit",
                                       bound=Attribute)


@irdl_attr_definition
class UnrankedMemrefType(Generic[_UnrankedMemrefTypeElems],
                         ParametrizedAttribute, TypeAttribute):
    name = "unranked_memref"

    element_type: ParameterDef[_UnrankedMemrefTypeElems]
    memory_space: ParameterDef[Attribute]

    @staticmethod
    def from_type(
        referenced_type: _UnrankedMemrefTypeElemsInit,
        memory_space: Attribute = NoneAttr(),
    ) -> UnrankedMemrefType[_UnrankedMemrefTypeElemsInit]:
        return UnrankedMemrefType([referenced_type, memory_space])


AnyUnrankedMemrefType: TypeAlias = UnrankedMemrefType[Attribute]


@irdl_op_definition
class Load(Operation):
    name = "memref.load"
    memref: Annotated[Operand, MemRefType[Attribute]]
    indices: Annotated[VarOperand, IndexType]
    res: Annotated[OpResult, AnyAttr()]

    # TODO varargs for indexing, which must match the memref dimensions
    # Problem: memref dimensions require variadic type parameters,
    # which is subject to change

    def verify_(self):
        if not isinstance(self.memref.typ, MemRefType):
            raise VerifyException("expected a memreftype")

        memref_typ = cast(MemRefType[Attribute], self.memref.typ)

        if memref_typ.element_type != self.res.typ:
            raise Exception(
                "expected return type to match the MemRef element type")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise Exception("expected an index for each dimension")

    @staticmethod
    def get(ref: SSAValue | Operation,
            indices: Sequence[SSAValue | Operation]) -> Load:
        ssa_value = SSAValue.get(ref)
        typ = ssa_value.typ
        typ = cast(MemRefType[Attribute], typ)
        return Load.build(operands=[ref, indices],
                          result_types=[typ.element_type])


@irdl_op_definition
class Store(Operation):
    name = "memref.store"
    value: Annotated[Operand, AnyAttr()]
    memref: Annotated[Operand, MemRefType[Attribute]]
    indices: Annotated[VarOperand, IndexType]

    def verify_(self):
        if not isinstance(self.memref.typ, MemRefType):
            raise VerifyException("expected a memreftype")

        memref_typ = cast(MemRefType[Attribute], self.memref.typ)

        if memref_typ.element_type != self.value.typ:
            raise Exception(
                "Expected value type to match the MemRef element type")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise Exception("Expected an index for each dimension")

    @staticmethod
    def get(value: Operation | SSAValue, ref: Operation | SSAValue,
            indices: Sequence[Operation | SSAValue]) -> Store:
        return Store.build(operands=[value, ref, indices])


@irdl_op_definition
class Alloc(Operation):
    name = "memref.alloc"

    dynamic_sizes: Annotated[VarOperand, IndexType]
    symbol_operands: Annotated[VarOperand, IndexType]

    memref: Annotated[OpResult, MemRefType[Attribute]]

    # TODO how to constraint the IntegerAttr type?
    alignment: OpAttr[AnyIntegerAttr]

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(return_type: Attribute,
            alignment: int,
            shape: Optional[List[int | AnyIntegerAttr]] = None) -> Alloc:
        if shape is None:
            shape = [1]
        return Alloc.build(operands=[[], []],
                           result_types=[
                               MemRefType.from_element_type_and_shape(
                                   return_type, shape)
                           ],
                           attributes={
                               "alignment":
                               IntegerAttr.from_int_and_width(alignment, 64)
                           })


@irdl_op_definition
class Alloca(Operation):
    name = "memref.alloca"

    dynamic_sizes: Annotated[VarOperand, IndexType]
    symbol_operands: Annotated[VarOperand, IndexType]

    memref: Annotated[OpResult, MemRefType[Attribute]]

    # TODO how to constraint the IntegerAttr type?
    alignment: OpAttr[AnyIntegerAttr]

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(return_type: Attribute,
            alignment: int,
            shape: Optional[List[int | AnyIntegerAttr]] = None,
            dynamic_sizes: list[SSAValue | Operation] | None = None) -> Alloca:
        if shape is None:
            shape = [1]

        if dynamic_sizes is None:
            dynamic_sizes = []

        return Alloca.build(operands=[dynamic_sizes, []],
                            result_types=[
                                MemRefType.from_element_type_and_shape(
                                    return_type, shape)
                            ],
                            attributes={
                                "alignment":
                                IntegerAttr.from_int_and_width(alignment, 64)
                            })


@irdl_op_definition
class Dealloc(Operation):
    name = "memref.dealloc"
    memref: Annotated[Operand,
                      MemRefType[Attribute] | UnrankedMemrefType[Attribute]]

    @staticmethod
    def get(operand: Operation | SSAValue) -> Dealloc:
        return Dealloc.build(operands=[operand])


@irdl_op_definition
class Dealloca(Operation):
    name = "memref.dealloca"
    memref: Annotated[Operand, MemRefType[Attribute]]

    @staticmethod
    def get(operand: Operation | SSAValue) -> Dealloca:
        return Dealloca.build(operands=[operand])


@irdl_op_definition
class GetGlobal(Operation):
    name = "memref.get_global"
    memref: Annotated[OpResult, MemRefType[Attribute]]

    def verify_(self) -> None:
        if 'name' not in self.attributes:
            raise VerifyException("GetGlobal requires a 'name' attribute")

        if not isinstance(self.attributes['name'], SymbolRefAttr):
            raise VerifyException(
                "expected 'name' attribute to be a SymbolRefAttr")

    @staticmethod
    def get(name: str, return_type: Attribute) -> GetGlobal:
        return GetGlobal.build(result_types=[return_type],
                               attributes={"name": SymbolRefAttr.build(name)})

    # TODO how to verify the types, as the global might be defined in another
    # compilation unit


@irdl_op_definition
class Global(Operation):
    name = "memref.global"

    sym_name: OpAttr[StringAttr]
    sym_visibility: OpAttr[StringAttr]
    type: OpAttr[Attribute]
    initial_value: OpAttr[Attribute]

    def verify_(self) -> None:
        if not isinstance(self.type, MemRefType):
            raise Exception("Global expects a MemRefType")

        if not isinstance(self.initial_value,
                          UnitAttr | DenseIntOrFPElementsAttr):
            raise Exception("Global initial value is expected to be a "
                            "dense type or an unit attribute")

    @staticmethod
    def get(
        sym_name: StringAttr,
        typ: Attribute,
        initial_value: Attribute,
        sym_visibility: StringAttr = StringAttr("private")
    ) -> Global:
        return Global.build(
            attributes={
                "sym_name": sym_name,
                "type": typ,
                "initial_value": initial_value,
                "sym_visibility": sym_visibility
            })


@irdl_op_definition
class Dim(Operation):
    name = "memref.dim"

    source: Annotated[Operand,
                      MemRefType[Attribute] | UnrankedMemrefType[Attribute]]
    index: Annotated[Operand, IndexType]

    result: Annotated[OpResult, IndexType]

    @classmethod
    def from_source_and_index(cls, source: SSAValue | Operation,
                              index: SSAValue | Operation):
        return cls.build(operands=[source, index], result_types=[IndexType()])


@irdl_op_definition
class Rank(Operation):
    name = "memref.rank"

    source: Annotated[Operand, MemRefType[Attribute]]

    rank: Annotated[OpResult, IndexType]

    @classmethod
    def from_memref(cls, memref: Operation | SSAValue):
        return cls.build(operands=[memref], result_types=[IndexType()])


@irdl_op_definition
class ExtractAlignedPointerAsIndexOp(Operation):
    name = "memref.extract_aligned_pointer_as_index"

    source: Annotated[Operand, MemRefType]

    aligned_pointer: Annotated[OpResult, IndexType]

    @staticmethod
    def get(source: SSAValue | Operation):
        return ExtractAlignedPointerAsIndexOp.build(operands=[source],
                                                    result_types=[IndexType()])


@irdl_op_definition
class Subview(Operation):
    name = "memref.subview"

    source: Annotated[Operand, MemRefType]
    offsets: Annotated[VarOperand, IndexType]
    sizes: Annotated[VarOperand, IndexType]
    strides: Annotated[VarOperand, IndexType]
    static_offsets: OpAttr[DenseArrayBase]
    static_sizes: OpAttr[DenseArrayBase]
    static_strides: OpAttr[DenseArrayBase]
    result: Annotated[OpResult, MemRefType]

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def from_static_parameters(
        source: SSAValue | Operation,
        source_element_type: Attribute,
        source_shape: Sequence[int],
        offsets: Sequence[int],
        sizes: Sequence[int],
        strides: Sequence[int],
    ) -> Subview:

        source = SSAValue.get(source)

        layout_strides = [1]
        for input_size in reversed(source_shape[1:]):
            layout_strides.insert(0, layout_strides[0] * input_size)

        layout_offset = sum(stride * offset
                            for stride, offset in zip(layout_strides, offsets))

        for i in range(len(layout_strides)):
            layout_strides[i] *= strides[i]

        layout = StridedLayoutAttr(layout_strides, layout_offset)

        return_typ = MemRefType.from_element_type_and_shape(
            source_element_type,
            sizes,
            layout,
        )

        return Subview.build(operands=[source, [], [], []],
                             result_types=[return_typ],
                             attributes={
                                 "static_offsets":
                                 DenseArrayBase.from_list(i64, offsets),
                                 "static_sizes":
                                 DenseArrayBase.from_list(i64, sizes),
                                 "static_strides":
                                 DenseArrayBase.from_list(i64, strides)
                             })


@irdl_op_definition
class Cast(Operation):
    name = "memref.cast"

    source: Annotated[Operand, MemRefType | UnrankedMemrefType]
    dest: Annotated[OpResult, MemRefType | UnrankedMemrefType]

    @staticmethod
    def get(source: SSAValue | Operation,
            type: MemRefType[Attribute] | UnrankedMemrefType[Attribute]):
        return Cast.build(operands=[source], result_types=[type])


MemRef = Dialect([
    Load,
    Store,
    Alloc,
    Alloca,
    Dealloc,
    GetGlobal,
    Global,
    Dim,
    ExtractAlignedPointerAsIndexOp,
    Subview,
    Cast,
], [MemRefType, UnrankedMemrefType])



from __future__ import annotations
from typing import Annotated, Generic, Sequence, Type, TypeVar

from xdsl.ir import (Attribute, TypeAttribute, OpResult, Operation, Dialect,
                     ParametrizedAttribute, Region, SSAValue)
from xdsl.irdl import (AttrSizedOperandSegments, Operand, OptOpAttr,
                       OptOpResult, OptOperand, ParameterDef, VarOperand,
                       irdl_op_definition, irdl_attr_definition,
                       SingleBlockRegion, OpAttr)
from xdsl.dialects.builtin import IndexType, StringAttr, SymbolRefAttr, UnitAttr, i32
from xdsl.dialects import memref
from xdsl.parser import BaseParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException


@irdl_attr_definition
class AsyncTokenType(ParametrizedAttribute, TypeAttribute):
    name = "gpu.async.token"


@irdl_attr_definition
class _AllReduceOperationAttr(ParametrizedAttribute):
    name = "all_reduce_op"

    param: ParameterDef[StringAttr]

    def print_parameters(self, printer: Printer) -> None:
        printer.print(f"all_reduce_op {self.param.data}")


@irdl_attr_definition
class _DimensionAttr(ParametrizedAttribute):
    name = "dim"

    param: ParameterDef[StringAttr]

    def print_parameters(self, printer: Printer) -> None:
        printer.print(f"dim {self.param.data}")


T = TypeVar('T',
            bound=_AllReduceOperationAttr | _DimensionAttr,
            covariant=True)


@irdl_attr_definition
class _GPUAttr(ParametrizedAttribute, Generic[T]):
    name = "gpu"

    value: ParameterDef[T]

    @staticmethod
    def parse_parameters(parser: BaseParser) -> list[Attribute]:
        parser.parse_characters(
            "<",
            f"Expected <. gpu attributes currently have the #gpu<name value> syntax."
        )
        ntok = parser.tokenizer.next_token()

        if ntok.text == "dim":
            attrtype = _DimensionAttr
            vtok = parser.tokenizer.next_token()
            if vtok.text not in ["x", "y", "z"]:
                parser.raise_error(
                    f"Unexpected dim {vtok.text}. A gpu dim can only be x, y, or z",
                    vtok)

        elif ntok.text == "all_reduce_op":
            attrtype = _AllReduceOperationAttr
            vtok = parser.tokenizer.next_token()
            if vtok.text not in [
                    "add", "and", "max", "min", "mul", "or", "xor"
            ]:
                parser.raise_error(
                    f"Unexpected op {vtok.text}. A gpu all_reduce_op can only be add, "
                    "and, max, min, mul, or, or xor ", vtok)
        else:
            parser.raise_error(
                f"Unexpected token {ntok.text}. Expected dim or all_reduce_op",
                ntok)
        parser.parse_characters(
            ">",
            f"Expected >. gpu attributes currently have the #gpu<name value> syntax."
        )
        return [attrtype([StringAttr(vtok.text)])]

    @classmethod
    def from_op(cls: Type[_GPUAttr[_AllReduceOperationAttr]],
                value: str) -> AllReduceOperationAttr:
        return AllReduceOperationAttr(
            [_AllReduceOperationAttr([StringAttr(value)])])

    @property
    def data(self) -> str:
        return self.value.param.data

    @classmethod
    def from_dimension(cls: Type[_GPUAttr[_DimensionAttr]],
                       value: str) -> DimensionAttr:
        return DimensionAttr([_DimensionAttr([StringAttr(value)])])

    def print_parameters(self, printer: Printer) -> None:
        printer.print_string("<")
        self.value.print_parameters(printer)
        printer.print_string(">")


DimensionAttr = _GPUAttr[_DimensionAttr]
AllReduceOperationAttr = _GPUAttr[_AllReduceOperationAttr]

_Element = TypeVar("_Element", bound=Attribute, covariant=True)


@irdl_op_definition
class AllocOp(Operation):
    name = "gpu.alloc"
    hostShared: OptOpAttr[UnitAttr]
    asyncDependencies: Annotated[VarOperand, AsyncTokenType]
    dynamicSizes: Annotated[VarOperand, IndexType]
    symbolOperands: Annotated[VarOperand, IndexType]

    irdl_options = [AttrSizedOperandSegments()]

    result: Annotated[OpResult, memref.MemRefType[Attribute]]
    asyncToken: Annotated[OptOpResult, AsyncTokenType]

    def verify_(self) -> None:
        ndyn = len(self.dynamicSizes)
        assert isinstance(self.result.typ, memref.MemRefType)
        typ: memref.MemRefType[Attribute] = self.result.typ
        ndyn_typ = len([i for i in typ.shape.data if i.value.data == -1])
        if ndyn != ndyn_typ:
            raise VerifyException(
                f"Expected {ndyn_typ} dynamic sizes, got {ndyn}. All "
                "dynamic sizes need to be set in the alloc operation.")

    @staticmethod
    def get(return_type: memref.MemRefType[_Element],
            dynamic_sizes: Sequence[SSAValue | Operation]
            | None = None,
            host_shared: bool = False,
            async_dependencies: Sequence[SSAValue | Operation] | None = None,
            is_async: bool = False) -> AllocOp:
        token_return = [AsyncTokenType()] if is_async else []
        dynamic_sizes_vals: list[SSAValue] = [
            SSAValue.get(e) for e in dynamic_sizes
        ] if dynamic_sizes else []
        async_dependencies_vals: list[SSAValue] = [
            SSAValue.get(e) for e in async_dependencies
        ] if async_dependencies else []
        attributes: dict[str, Attribute] = {
            "hostShared": UnitAttr()
        } if host_shared else {}
        return AllocOp.build(
            operands=[async_dependencies_vals, dynamic_sizes_vals, []],
            result_types=[return_type, token_return],
            attributes=attributes)


@irdl_op_definition
class AllReduceOp(Operation):
    name = "gpu.all_reduce"
    op: OptOpAttr[AllReduceOperationAttr]
    uniform: OptOpAttr[UnitAttr]
    operand: Annotated[Operand, Attribute]
    result: Annotated[OpResult, Attribute]
    body: Region

    @staticmethod
    def from_op(op: AllReduceOperationAttr,
                operand: SSAValue | Operation,
                uniform: UnitAttr | None = None):

        return AllReduceOp.build(operands=[operand],
                                 result_types=[SSAValue.get(operand).typ],
                                 attributes={
                                     "op": op,
                                     "uniform": uniform,
                                 },
                                 regions=[Region()])

    @staticmethod
    def from_body(body: Region,
                  operand: SSAValue | Operation,
                  uniform: UnitAttr | None = None):
        return AllReduceOp.build(
            operands=[operand],
            result_types=[SSAValue.get(operand).typ],
            attributes={"uniform": uniform} if uniform is not None else {},
            regions=[body])

    def verify_(self) -> None:
        if self.result.typ != self.operand.typ:
            raise VerifyException(
                f"Type mismatch: result type is {self.result.typ}, operand type is "
                f"{self.operand.typ}. They must be the same type for gpu.all_reduce"
            )

        non_empty_body = any([len(b.ops) > 0 for b in self.body.blocks])
        op_attr = self.op is not None
        if non_empty_body == op_attr:
            if op_attr:
                raise VerifyException(
                    f"gpu.all_reduce can't have both a non-empty region and an op "
                    "attribute.")
            else:
                raise VerifyException(
                    f"gpu.all_reduce need either a non empty body or an op attribute."
                )
        if non_empty_body:
            region_args = self.body.blocks[0].args
            args_types = [r.typ for r in region_args]
            if args_types != [self.result.typ, self.operand.typ]:
                raise VerifyException(
                    f"Expected {[str(t) for t in [self.result.typ, self.operand.typ]]}, "
                    f"got {[str(t) for t in args_types]}. A gpu.all_reduce's body must "
                    "have two arguments matching the result type.")


@irdl_op_definition
class BarrierOp(Operation):
    name = "gpu.barrier"

    @staticmethod
    def get() -> BarrierOp:
        return BarrierOp.build()


@irdl_op_definition
class BlockDimOp(Operation):
    name = "gpu.block_dim"
    dimension: OpAttr[DimensionAttr]
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get(dim: DimensionAttr) -> BlockDimOp:
        return BlockDimOp.build(result_types=[IndexType()],
                                attributes={"dimension": dim})


@irdl_op_definition
class BlockIdOp(Operation):
    name = "gpu.block_id"
    dimension: OpAttr[DimensionAttr]
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get(dim: DimensionAttr) -> BlockIdOp:
        return BlockIdOp.build(result_types=[IndexType()],
                               attributes={"dimension": dim})


@irdl_op_definition
class DeallocOp(Operation):
    name = "gpu.dealloc"

    asyncDependencies: Annotated[VarOperand, AsyncTokenType]
    buffer: Annotated[Operand, memref.MemRefType]

    irdl_options = [AttrSizedOperandSegments()]

    asyncToken: Annotated[OptOpResult, AsyncTokenType]

    @staticmethod
    def get(buffer: SSAValue | Operation,
            async_dependencies: Sequence[SSAValue | Operation] | None = None,
            is_async: bool = False) -> DeallocOp:
        return DeallocOp.build(
            operands=[async_dependencies, buffer],
            result_types=[[AsyncTokenType()] if is_async else []])


@irdl_op_definition
class MemcpyOp(Operation):
    name = "gpu.memcpy"

    asyncDependencies: Annotated[VarOperand, AsyncTokenType]
    src: Annotated[Operand, memref.MemRefType]
    dst: Annotated[Operand, memref.MemRefType]

    irdl_options = [AttrSizedOperandSegments()]

    asyncToken: Annotated[OptOpResult, AsyncTokenType]

    @staticmethod
    def get(source: SSAValue | Operation,
            destination: SSAValue | Operation,
            async_dependencies: Sequence[SSAValue | Operation] | None = None,
            is_async: bool = False) -> MemcpyOp:
        return MemcpyOp.build(
            operands=[async_dependencies, source, destination],
            result_types=[[AsyncTokenType()] if is_async else []])

    def verify_(self) -> None:
        if self.src.typ != self.dst.typ:
            raise VerifyException(
                f"Expected {self.src.typ}, got {self.dst.typ}. gpu.memcpy source and "
                "destination types must match.")


@irdl_op_definition
class ModuleOp(Operation):
    name = "gpu.module"

    body: SingleBlockRegion
    sym_name: OpAttr[StringAttr]

    @staticmethod
    def get(name: SymbolRefAttr, ops: list[Operation]) -> ModuleOp:
        op = ModuleOp.build(attributes={"sym_name": name}, regions=[ops])
        return op

    def verify_(self):
        if (len(self.body.ops) == 0
                or not isinstance(self.body.ops[-1], ModuleEndOp)):
            raise VerifyException("gpu.module must end with gpu.module_end")


@irdl_op_definition
class GlobalIdOp(Operation):
    name = "gpu.global_id"
    dimension: OpAttr[DimensionAttr]
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get(dim: DimensionAttr) -> GlobalIdOp:
        return GlobalIdOp.build(result_types=[IndexType()],
                                attributes={"dimension": dim})


@irdl_op_definition
class GridDimOp(Operation):
    name = "gpu.grid_dim"
    dimension: OpAttr[DimensionAttr]
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get(dim: DimensionAttr) -> GridDimOp:
        return GridDimOp.build(result_types=[IndexType()],
                               attributes={"dimension": dim})


@irdl_op_definition
class HostRegisterOp(Operation):
    """
    This op maps the provided host buffer into the device address space.

    This operation may not be supported in every environment, there is not yet a way to
    check at runtime whether this feature is supported.
    Writes from the host are guaranteed to be visible to device kernels that are launched
    afterwards. Writes from the device are guaranteed to be visible on the host after
    synchronizing with the device kernel completion.
    """
    name = "gpu.host_register"

    value: Annotated[Operand, memref.UnrankedMemrefType]

    @staticmethod
    def from_memref(memref: SSAValue | Operation):
        return HostRegisterOp.build(operands=[SSAValue.get(memref)])


@irdl_op_definition
class LaneIdOp(Operation):
    name = "gpu.lane_id"
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get() -> LaneIdOp:
        return LaneIdOp.build(result_types=[IndexType()])


@irdl_op_definition
class LaunchOp(Operation):
    name = "gpu.launch"
    asyncDependencies: Annotated[VarOperand, AsyncTokenType]
    gridSizeX: Annotated[Operand, IndexType]
    gridSizeY: Annotated[Operand, IndexType]
    gridSizeZ: Annotated[Operand, IndexType]
    blockSizeX: Annotated[Operand, IndexType]
    blockSizeY: Annotated[Operand, IndexType]
    blockSizeZ: Annotated[Operand, IndexType]
    dynamicSharedMemorySize: Annotated[OptOperand, i32]
    asyncToken: Annotated[OptOpResult, AsyncTokenType]
    body: Region
    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(
        body: Region,
        gridSize: Sequence[SSAValue | Operation],
        blockSize: Sequence[SSAValue | Operation],
        async_launch: bool = False,
        asyncDependencies: Sequence[SSAValue | Operation] | None = None,
        dynamicSharedMemorySize: SSAValue | Operation | None = None
    ) -> LaunchOp:
        if len(gridSize) != 3:
            raise ValueError(
                f"LaunchOp must have 3 gridSizes, got {len(gridSize)}")
        if len(blockSize) != 3:
            raise ValueError(
                f"LaunchOp must have 3 blockSizes, got {len(blockSize)}")
        operands = [[] if asyncDependencies is None else
                    [SSAValue.get(a) for a in asyncDependencies]]

        operands += [SSAValue.get(gs) for gs in gridSize]
        operands += [SSAValue.get(bs) for bs in blockSize]
        operands += [[] if dynamicSharedMemorySize is None else
                     [SSAValue.get(dynamicSharedMemorySize)]]
        return LaunchOp.build(
            operands=operands,
            result_types=[[AsyncTokenType()] if async_launch else []],
            regions=[body])

    def verify_(self) -> None:
        if len(self.body.blocks) == 0 or all(
            [len(b.ops) == 0 for b in self.body.blocks]):
            raise VerifyException("gpu.launch requires a non-empty body.")
        body_args = self.body.blocks[0].args
        args_type = [a.typ for a in body_args]
        if args_type != [IndexType()] * 12:
            raise VerifyException(
                f"Expected [12 x {str(IndexType())}], got {[str(t) for t in args_type]}. "
                "gpu.launch's body arguments are 12 index arguments, with 3 block "
                "indices, 3 block sizes, 3 thread indices, and 3 thread counts"
            )


@irdl_op_definition
class ModuleEndOp(Operation):
    name = "gpu.module_end"

    @staticmethod
    def get() -> ModuleEndOp:
        return ModuleEndOp.build()


@irdl_op_definition
class NumSubgroupsOp(Operation):
    name = "gpu.num_subgroups"
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get() -> NumSubgroupsOp:
        return NumSubgroupsOp.build(result_types=[IndexType()])


@irdl_op_definition
class SetDefaultDeviceOp(Operation):
    name = "gpu.set_default_device"
    devIndex: Annotated[Operand, i32]

    @staticmethod
    def get(devIndex: SSAValue | Operation) -> SetDefaultDeviceOp:
        return SetDefaultDeviceOp.build(operands=[SSAValue.get(devIndex)])


@irdl_op_definition
class SubgroupIdOp(Operation):
    name = "gpu.subgroup_id"
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get() -> SubgroupIdOp:
        return SubgroupIdOp.build(result_types=[IndexType()])


@irdl_op_definition
class SubgroupSizeOp(Operation):
    name = "gpu.subgroup_size"
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get() -> SubgroupSizeOp:
        return SubgroupSizeOp.build(result_types=[IndexType()])


@irdl_op_definition
class TerminatorOp(Operation):
    name = "gpu.terminator"

    @staticmethod
    def get() -> TerminatorOp:
        return TerminatorOp.build()

    def verify_(self) -> None:
        block = self.parent_block()
        op = self.parent_op()
        if block is not None:
            if self is not block.ops[-1]:
                raise VerifyException(
                    "A gpu.terminator must terminate its parent block")
        if op is not None and not isinstance(op, LaunchOp):
            raise VerifyException(
                "gpu.terminator is only meant to terminate gpu.launch")


@irdl_op_definition
class ThreadIdOp(Operation):
    name = "gpu.thread_id"
    dimension: OpAttr[DimensionAttr]
    result: Annotated[OpResult, IndexType]

    @staticmethod
    def get(dim: DimensionAttr) -> ThreadIdOp:
        return ThreadIdOp.build(result_types=[IndexType()],
                                attributes={"dimension": dim})


@irdl_op_definition
class YieldOp(Operation):
    name = "gpu.yield"
    values: Annotated[VarOperand, Attribute]

    @staticmethod
    def get(operands: list[SSAValue | Operation]) -> YieldOp:
        return YieldOp.build([operands])

    def verify_(self) -> None:
        block = self.parent_block()
        op = self.parent_op()
        if block is not None:
            if self is not block.ops[-1]:
                raise VerifyException(
                    "A gpu.yield must terminate its parent block")
        if op is not None:
            yield_type = [o.typ for o in self.values]
            result_type = [r.typ for r in op.results]
            if yield_type != result_type:
                raise VerifyException(
                    f"Expected {[str(t) for t in result_type]}, got {[str(t) for t in yield_type]}. The gpu.yield values "
                    "types must match its enclosing operation result types.")


#_GPUAttr has to be registered instead of DimensionAttr and AllReduceOperationAttr here.
# This is a hack to fit MLIR's syntax in xDSL's way of parsing attributes, without making GPU builtin.
# Hopefully MLIR will parse it in a more xDSL-friendly way soon, so all that can be factored in proper xDSL
# atrributes.
GPU = Dialect([
    AllocOp,
    AllReduceOp,
    BarrierOp,
    BlockDimOp,
    BlockIdOp,
    DeallocOp,
    GlobalIdOp,
    GridDimOp,
    HostRegisterOp,
    LaneIdOp,
    LaunchOp,
    MemcpyOp,
    ModuleOp,
    ModuleEndOp,
    NumSubgroupsOp,
    SetDefaultDeviceOp,
    SubgroupIdOp,
    SubgroupSizeOp,
    TerminatorOp,
    ThreadIdOp,
    YieldOp,
], [_GPUAttr])



from __future__ import annotations

from abc import ABC
from enum import Enum
from typing import cast, TypeVar, Generic, Sequence

from xdsl.utils.hints import isa
from xdsl.dialects import llvm
from xdsl.dialects.builtin import (IntegerType, Signedness, StringAttr,
                                   AnyFloat, i32)
from xdsl.dialects.memref import MemRefType
from xdsl.ir import (Operation, Attribute, SSAValue, OpResult,
                     ParametrizedAttribute, Dialect, TypeAttribute)
from xdsl.irdl import (Operand, Annotated, irdl_op_definition,
                       irdl_attr_definition, OpAttr, OptOpResult, ParameterDef,
                       OptOperand, OptOpAttr)

t_bool: IntegerType = IntegerType(1, Signedness.SIGNLESS)

AnyNumericType = AnyFloat | IntegerType


@irdl_attr_definition
class OperationType(ParametrizedAttribute, TypeAttribute):
    """
    This type represents the MPI_Op type.

    They are used by the reduction MPI functions
    """
    name = 'mpi.operation'

    op_str: ParameterDef[StringAttr]


class MpiOp:
    """
    A collection of MPI_Op types used for
    """
    MPI_MAX = OperationType([StringAttr("MPI_MAX")])
    MPI_MIN = OperationType([StringAttr("MPI_MIN")])
    MPI_SUM = OperationType([StringAttr("MPI_SUM")])
    MPI_PROD = OperationType([StringAttr("MPI_PROD")])
    MPI_LAND = OperationType([StringAttr("MPI_LAND")])
    MPI_BAND = OperationType([StringAttr("MPI_BAND")])
    MPI_LOR = OperationType([StringAttr("MPI_LOR")])
    MPI_BOR = OperationType([StringAttr("MPI_BOR")])
    MPI_LXOR = OperationType([StringAttr("MPI_LXOR")])
    MPI_BXOR = OperationType([StringAttr("MPI_BXOR")])
    MPI_MINLOC = OperationType([StringAttr("MPI_MINLOC")])
    MPI_MAXLOC = OperationType([StringAttr("MPI_MAXLOC")])
    MPI_REPLACE = OperationType([StringAttr("MPI_REPLACE")])
    MPI_NO_OP = OperationType([StringAttr("MPI_NO_OP")])


@irdl_attr_definition
class RequestType(ParametrizedAttribute, TypeAttribute):
    """
    This type represents the MPI_Request type.

    They are used by the asynchronous MPI functions
    """
    name = 'mpi.request'


@irdl_attr_definition
class StatusType(ParametrizedAttribute, TypeAttribute):
    """
    This type represents the MPI_Status type.

    It's a struct containing status information for requests.
    """
    name = 'mpi.status'


@irdl_attr_definition
class DataType(ParametrizedAttribute, TypeAttribute):
    """
    This type represents MPI_Datatype
    """
    name = 'mpi.datatype'


VectorWrappable = RequestType | StatusType | DataType
_VectorT = TypeVar('_VectorT', bound=VectorWrappable)


@irdl_attr_definition
class VectorType(Generic[_VectorT], ParametrizedAttribute, TypeAttribute):
    """
    This type holds multiple MPI types
    """
    name = 'mpi.vector'
    wrapped_type: ParameterDef[_VectorT]

    @staticmethod
    def of(typ: type[_VectorT]) -> VectorType[_VectorT]:
        return VectorType([typ([])])


class StatusTypeField(Enum):
    """
    This enum lists all fields in the MPI_Status struct
    """
    MPI_SOURCE = 'MPI_SOURCE'
    MPI_TAG = 'MPI_TAG'
    MPI_ERROR = 'MPI_ERROR'


class MPIBaseOp(Operation, ABC):
    """
    Base class for MPI Operations
    """
    pass


@irdl_op_definition
class Reduce(MPIBaseOp):
    """
    This wraps the MPI_Reduce function (blocking reduction)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Reduce.html

    ## The MPI_Reduce Function Docs:

    int MPI_Reduce(const void *sendbuf, void *recvbuf, int count,
               MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)

        sendbuf: address of send buffer (choice)
        recvbuf: address of receive buffer (choice)
        count: number of elements in send buffer (non-negative integer)
        datatype: data type of elements of send buffer (handle)
        op: reduce operation (handle)
        root: rank of root process (integer)
        comm: communicator (handle)

    ## Our Abstraction:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = "mpi.reduce"

    send_buffer: Annotated[Operand, Attribute]
    recv_buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    operationtype: OpAttr[OperationType]
    root: Annotated[Operand, i32]

    @staticmethod
    def get(
        send_buffer: SSAValue | Operation,
        recv_buffer: SSAValue | Operation,
        count: SSAValue | Operation,
        datatype: SSAValue | Operation,
        operationtype: OperationType,
        root: SSAValue | Operation,
    ):
        return Reduce.build(
            operands=[send_buffer, recv_buffer, count, datatype, root],
            attributes={"operationtype": operationtype},
            result_types=[],
        )


@irdl_op_definition
class Allreduce(MPIBaseOp):
    """
    This wraps the MPI_Allreduce function (blocking all reduction)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Allreduce.html

    ## The MPI_Allreduce Function Docs:

    int MPI_Allreduce(const void *sendbuf, void *recvbuf, int count,
                  MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)

        sendbuf: address of send buffer (choice)
        recvbuf: address of receive buffer (choice)
        count: number of elements in send buffer (non-negative integer)
        datatype: data type of elements of send buffer (handle)
        op: reduce operation (handle)
        comm: communicator (handle)

    ## Our Abstraction:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = "mpi.allreduce"

    recv_buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    send_buffer: Annotated[OptOperand, Attribute]
    operationtype: OpAttr[OperationType]

    @staticmethod
    def get(
        send_buffer: SSAValue | Operation | None,
        recv_buffer: SSAValue | Operation,
        count: SSAValue | Operation,
        datatype: SSAValue | Operation,
        operationtype: OperationType,
    ):

        operands_to_add: Sequence[SSAValue | Operation
                                  | Sequence[SSAValue | Operation]] = []
        if send_buffer is None:
            operands_to_add = [recv_buffer, count, datatype, []]
        else:
            operands_to_add = [recv_buffer, count, datatype, [send_buffer]]

        return Allreduce.build(
            operands=operands_to_add,
            attributes={"operationtype": operationtype},
            result_types=[],
        )


@irdl_op_definition
class Bcast(MPIBaseOp):
    """
    This wraps the MPI_Bcast function (blocking broadcast)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Bcast.html

    ## The MPI_Bcast Function Docs:

    int MPI_Bcast(void *buffer, int count, MPI_Datatype datatype, int root,
              MPI_Comm comm)

        buffer: starting address of buffer (choice)
        count: number of elements in send buffer (non-negative integer)
        datatype: data type of elements of send buffer (handle)
        root: rank of broadcast root (integer)
        comm: communicator (handle)

    ## Our Abstraction:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = "mpi.bcast"

    buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    root: Annotated[Operand, i32]

    @staticmethod
    def get(
        buffer: SSAValue | Operation,
        count: SSAValue | Operation,
        datatype: SSAValue | Operation,
        root: SSAValue | Operation,
    ):
        return Bcast.build(
            operands=[buffer, count, datatype, root],
            result_types=[],
        )


@irdl_op_definition
class Isend(MPIBaseOp):
    """
    This wraps the MPI_Isend function (nonblocking send)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Isend.html

    ## The MPI_Isend Function Docs:

    int MPI_Isend(const void *buf, int count, MPI_Datatype datatype, int dest,
         int tag, MPI_Comm comm, MPI_Request *request)

        - buf: Initial address of send buffer (choice).
        - count: Number of elements in send buffer (integer).
        - datatype: Datatype of each send buffer element (handle).
        - dest: Rank of destination (integer).
        - tag: Message tag (integer).
        - comm: Communicator (handle).

    ## Our Abstraction:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = 'mpi.isend'

    buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    dest: Annotated[Operand, i32]
    tag: Annotated[Operand, i32]
    request: Annotated[Operand, RequestType]

    @staticmethod
    def get(
        buffer: SSAValue | Operation,
        count: SSAValue | Operation,
        datatype: SSAValue | Operation,
        dest: SSAValue | Operation,
        tag: SSAValue | Operation,
        request: SSAValue | Operation,
    ):
        return Isend.build(
            operands=[buffer, count, datatype, dest, tag, request],
            result_types=[],
        )


@irdl_op_definition
class Send(MPIBaseOp):
    """
    This wraps the MPI_Send function (blocking send)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Send.html

    ## The MPI_Send Function Docs:

    int MPI_Send(const void *buf, int count, MPI_Datatype datatype, int dest,
             int tag, MPI_Comm comm)

        - buf: Initial address of send buffer (choice).
        - count: Number of elements in send buffer (non-negative integer).
        - datatype: Datatype of each send buffer element (handle).
        - dest: Rank of destination (integer).
        - tag: Message tag (integer).
        - comm: Communicator (handle).

    ## Our Abstraction:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = 'mpi.send'

    buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    dest: Annotated[Operand, i32]
    tag: Annotated[Operand, i32]

    @staticmethod
    def get(buffer: SSAValue | Operation, count: SSAValue | Operation,
            datatype: SSAValue | Operation, dest: SSAValue | Operation,
            tag: SSAValue | Operation) -> Send:
        return Send.build(operands=[buffer, count, datatype, dest, tag],
                          result_types=[])


@irdl_op_definition
class Irecv(MPIBaseOp):
    """
    This wraps the MPI_Irecv function (nonblocking receive).
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Irecv.html

    ## The MPI_Irecv Function Docs:

    int MPI_Irecv(void *buf, int count, MPI_Datatype datatype,
            int source, int tag, MPI_Comm comm, MPI_Request *request)

        - buf: Initial address of receive buffer (choice).
        - count: Number of elements in receive buffer (integer).
        - datatype: Datatype of each receive buffer element (handle).
        - source: Rank of source (integer).
        - tag: Message tag (integer).
        - comm: Communicator (handle).
        - request: Communication request (handle).

    ## Our Abstractions:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = "mpi.irecv"

    buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    source: Annotated[Operand, i32]
    tag: Annotated[Operand, i32]
    request: Annotated[Operand, RequestType]

    @staticmethod
    def get(
        buffer: SSAValue | Operation,
        count: SSAValue | Operation,
        datatype: SSAValue | Operation,
        source: SSAValue | Operation,
        tag: SSAValue | Operation,
        request: SSAValue | Operation,
    ):
        return Irecv.build(
            operands=[buffer, count, datatype, source, tag, request],
            result_types=[],
        )


@irdl_op_definition
class Recv(MPIBaseOp):
    """
    This wraps the MPI_Recv function (blocking receive).
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Recv.html

    ## The MPI_Recv Function Docs:

    int MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
             MPI_Comm comm, MPI_Status *status)

        - buf: Initial address of receive buffer (choice).
        - count: Number of elements in receive buffer (integer).
        - datatype: Datatype of each receive buffer element (handle).
        - source: Rank of source (integer).
        - tag: Message tag (integer).
        - comm: Communicator (handle).
        - status: status object (Status).

    ## Our Abstractions:

        - We omit the possibility of using multiple communicators, defaulting
          to MPI_COMM_WORLD
    """

    name = "mpi.recv"

    buffer: Annotated[Operand, Attribute]
    count: Annotated[Operand, i32]
    datatype: Annotated[Operand, DataType]
    source: Annotated[Operand, i32]
    tag: Annotated[Operand, i32]

    status: Annotated[OptOpResult, StatusType]

    @staticmethod
    def get(buffer: SSAValue | Operation,
            count: SSAValue | Operation,
            datatype: SSAValue | Operation,
            source: SSAValue | Operation,
            tag: SSAValue | Operation,
            ignore_status: bool = True):
        return Recv.build(
            operands=[buffer, count, datatype, source, tag],
            result_types=[[]] if ignore_status else [[StatusType()]])


@irdl_op_definition
class Test(MPIBaseOp):
    """
    Class for wrapping the MPI_Test function (test for completion of request)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Test.html

    ## The MPI_Test Function Docs:

    int MPI_Test(MPI_Request *request, int *flag, MPI_Status *status)

        - request: Communication request (handle)
        - flag: true if operation completed (logical)
        - status: Status object (Status)
    """

    name = "mpi.test"

    request: Annotated[Operand, RequestType]

    flag: Annotated[OpResult, t_bool]
    status: Annotated[OpResult, StatusType]

    @staticmethod
    def get(request: Operand):
        return Test.build(operands=[request],
                          result_types=[t_bool, StatusType()])


@irdl_op_definition
class Wait(MPIBaseOp):
    """
    Class for wrapping the MPI_Wait function (blocking wait for request)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Wait.html

    ## The MPI_Test Function Docs:

    int MPI_Wait(MPI_Request *request, MPI_Status *status)

        - request: Request (handle)
        - status: Status object (Status)
    """

    name = "mpi.wait"

    request: Annotated[Operand, RequestType]
    status: Annotated[OptOpResult, StatusType]

    @staticmethod
    def get(request: Operand, ignore_status: bool = True):
        result_types: list[list[Attribute]] = [[StatusType()]]
        if ignore_status:
            result_types = [[]]

        return Wait.build(operands=[request], result_types=result_types)


@irdl_op_definition
class Waitall(MPIBaseOp):
    """
    Class for wrapping the MPI_Waitall function (blocking wait for requests)
    https://www.mpich.org/static/docs/v4.1/www3/MPI_Waitall.html

    ## The MPI_Test Function Docs:

    int MPI_Waitall(int count, MPI_Request array_of_requests[],
                MPI_Status *array_of_statuses)

        - count: Number of handles
        - array_of_requests: Request handles
        - array_of_statuses: Status objects
    """

    name = "mpi.waitall"

    requests: Annotated[Operand, VectorType[RequestType]]
    count: Annotated[Operand, i32]
    statuses: Annotated[OptOpResult, VectorType[StatusType]]

    @staticmethod
    def get(requests: Operand, count: Operand, ignore_status: bool = True):
        result_types: list[list[Attribute]] = [[VectorType.of(StatusType)]]
        if ignore_status:
            result_types = [[]]

        return Waitall.build(operands=[requests, count],
                             result_types=result_types)


@irdl_op_definition
class GetStatusField(MPIBaseOp):
    """
    Accessors for the MPI_Status struct

    This allows access to the three integer properties in the struct called
        - MPI_SOURCE
        - MPI_TAG
        - MPI_ERROR

    All fields are of type int.
    """
    name = "mpi.status.get"

    status: Annotated[Operand, StatusType]

    field: OpAttr[StringAttr]

    result: Annotated[OpResult, i32]

    @staticmethod
    def get(status_obj: Operand, field: StatusTypeField):
        return GetStatusField.build(
            operands=[status_obj],
            attributes={'field': StringAttr(field.value)},
            result_types=[i32])


@irdl_op_definition
class CommRank(MPIBaseOp):
    """
    Represents the MPI_Comm_size(MPI_Comm comm, int *rank) function call which returns the rank of the communicator

    Currently limited to COMM_WORLD
    """
    name = "mpi.comm.rank"

    rank: Annotated[OpResult, i32]

    @staticmethod
    def get():
        return CommRank.build(result_types=[i32])


@irdl_op_definition
class CommSize(MPIBaseOp):
    """
    Represents the MPI_Comm_size(MPI_Comm comm, int *size) function call which returns the size of the communicator

    Currently limited to COMM_WORLD
    """
    name = "mpi.comm.size"

    size: Annotated[OpResult, i32]

    @staticmethod
    def get():
        return CommSize.build(result_types=[i32])


@irdl_op_definition
class Init(MPIBaseOp):
    """
    This represents a bare MPI_Init call with both args being nullptr
    """
    name = "mpi.init"


@irdl_op_definition
class Finalize(MPIBaseOp):
    """
    This represents an MPI_Finalize call with both args being nullptr
    """
    name = "mpi.finalize"


@irdl_op_definition
class UnwrapMemrefOp(MPIBaseOp):
    """
    This Op can be used as a helper to get memrefs into MPI calls.

    It takes any MemRef as input, and returns an llvm.ptr, element count and MPI_Datatype.
    """
    name = "mpi.unwrap_memref"

    ref: Annotated[Operand, MemRefType[AnyNumericType]]

    ptr: Annotated[OpResult, llvm.LLVMPointerType]
    len: Annotated[OpResult, i32]
    typ: Annotated[OpResult, DataType]

    @staticmethod
    def get(ref: SSAValue | Operation) -> UnwrapMemrefOp:
        ssa_val = SSAValue.get(ref)
        assert isinstance(ssa_val.typ, MemRefType)
        elem_typ = cast(MemRefType[AnyNumericType], ssa_val.typ).element_type

        return UnwrapMemrefOp.build(operands=[ref],
                                    result_types=[
                                        llvm.LLVMPointerType.typed(elem_typ),
                                        i32,
                                        DataType()
                                    ])


@irdl_op_definition
class GetDtypeOp(MPIBaseOp):
    """
    This op is used to convert MLIR types to MPI_Datatype constants.

    So, e.g. if you want to get the `MPI_Datatype` for an `i32` you can use

        %dtype = "mpi.get_dtype"() {"dtype" = i32} : () -> mpi.datatype

    to get the magic constant. See `_MPIToLLVMRewriteBase._translate_to_mpi_type`
    docstring for more detail on which types are supported.
    """
    name = "mpi.get_dtype"

    dtype: OpAttr[Attribute]

    result: Annotated[OpResult, DataType]

    @staticmethod
    def get(typ: Attribute):
        return GetDtypeOp.build(result_types=[DataType()],
                                attributes={'dtype': typ})


@irdl_op_definition
class AllocateTypeOp(MPIBaseOp):
    """
    This op is used to allocate a specific MPI dialect type with a set size, returning this
    in an MPI vector of that type

    This is useful as it means we can, in a self contained manner, store things like
    requests, statuses etc. It accepts the base type that the array will contain, the
    number of elements and an optional bindc_name which contains the name of the
    variable that this is allocating
    """
    name = "mpi.allocate"

    bindc_name: OptOpAttr[StringAttr]
    dtype: OpAttr[VectorWrappable]
    count: Annotated[Operand, i32]

    result: Annotated[OpResult, VectorType]

    @staticmethod
    def get(
        dtype: type[VectorWrappable],
        count: SSAValue | Operation,
        bindc_name: StringAttr | None = None,
    ) -> AllocateTypeOp:
        return AllocateTypeOp.build(result_types=[VectorType.of(dtype)],
                                    attributes={
                                        "dtype": dtype(),
                                        "bindc_name": bindc_name,
                                    },
                                    operands=[count])


@irdl_op_definition
class VectorGetOp(MPIBaseOp):
    """
    This op will retrieve an element of an MPI vector, it accepts the vector as
    an argument and the element index
    """
    name = "mpi.vector_get"

    vect: Annotated[Operand, VectorType]
    element: Annotated[Operand, i32]

    result: Annotated[OpResult, VectorWrappable]

    @staticmethod
    def get(vect: SSAValue | Operation,
            element: SSAValue | Operation) -> VectorGetOp:
        ssa_val = SSAValue.get(vect)
        assert isa(ssa_val.typ, VectorType[VectorWrappable])

        return VectorGetOp.build(result_types=[ssa_val.typ.wrapped_type],
                                 operands=[vect, element])


@irdl_op_definition
class NullRequestOp(MPIBaseOp):
    """
    This sets a given request object to the MPI_REQUEST_NULL magic
    value.

    Due to restrictions in the current MPI dialect, we can't return a
    new request object here. That will be fixed soon though!
    """
    name = "mpi.request_null"

    request: Annotated[Operand, RequestType]

    @staticmethod
    def get(req: SSAValue | Operation):
        return NullRequestOp.build(operands=[req])


MPI = Dialect([
    Isend,
    Irecv,
    Test,
    Recv,
    Send,
    Reduce,
    Allreduce,
    Bcast,
    Wait,
    GetStatusField,
    Init,
    Finalize,
    CommRank,
    UnwrapMemrefOp,
    GetDtypeOp,
    AllocateTypeOp,
    VectorGetOp,
], [
    OperationType,
    RequestType,
    StatusType,
    DataType,
    VectorType,
])
